/*

Geiger Mller Zhler fr Pollin Bausatz Best.Nr. 810 223

ATtiny2313 @ 1 MHz

*/

#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "main.h"
#include <util/delay.h>
#include "lcd.h"

#define LCD_LIGHT_ON  PORTD &= ~(1<<PD3);
#define LCD_LIGHT_OFF PORTD |=  (1<<PD3);
#define LIGHT_TIMEOUT 50                        // Timout fr LCD Beleuchtung in 1/10 Sekunden
#define BEEP_OFF  PORTB &= ~(1<<PB4);
#define BEEP_ON   PORTB |=  (1<<PB4);
#define BEEP_TIME 100                          // Zeit fr Pieper in Mikrosekunden
#define KEY_PRESSED  (!(PIND & (1<<PD6)))

volatile uint8_t flag_100ms;                    // Signal fuer Hauptschleife
volatile uint8_t time_ss, time_mm, time_hh;     // Uhrzeit
volatile uint16_t cpm, ca;                      // Pulse / Minute, Mittelwert
volatile uint32_t ct;                           // Gesamtanzahl Pulse
volatile uint16_t mittelwerte[4];
char tmp_string[11];

uint8_t user_char_0[8] PROGMEM ={
               // Durchschnitt Symbol
    0b00001,
    0b01110,
    0b10011,
    0b10101,
    0b11001,
    0b01110,
    0b10000,
    0b00000,
};


/*

Funktion zur Umwandlung einer vorzeichenlosen 32 Bit Zahl in einen String
 
*/
 
void my_uitoa(uint32_t zahl, char* string) {
  int8_t i;                             // schleifenzhler
 
  string[10]='\0';                       // String Terminator
  for(i=9; i>=0; i--) {
    string[i]=(zahl % 10) +'0';         // Modulo rechnen, dann den ASCII-Code von '0' addieren
    zahl /= 10;
  }
}

// Zahlenstring ausgeben, fhrende Nullen durch Leerzeichen ersetzen

void my_print_number(char* string) {
 
  uint8_t flag=0;       // Merker fr fhrende Nullen
  
  while(*string != '\0') {
    if (flag==1 || *string !='0') {
      lcd_data(*string);
      flag = 1;
    }
    else lcd_data(' ');         // Leerzeichen
    string++;
  }
}

// zweistellige Zahl mit fhrenden Nullen ausgeben

void print_time(uint8_t data) { 

    uint8_t a, b;

    a = data / 10;
    b = data % 10;
    lcd_data('0'+a);
    lcd_data('0'+b);
}

int main (void) {

    // lokale Kopien, fr atomaren Zugriff
    uint8_t m_time_ss, m_time_mm, m_time_hh;    // Uhrzeit
    uint16_t m_cpm, m_ca;                       // Pulse / Minute, Mittelwert
    uint32_t m_ct;                              // Gesamtanzahl Pulse
    
    // IO init

    DDRB = 0xFF;
    DDRD = ~((1<<PD6) | (1<<PD2));       // Taster, Pulseingang

    LCD_LIGHT_ON
    BEEP_OFF

    // Timer 1 init, mode 0, Teiler 1 = 1 MHz
    
    TCCR1A = 0; 
    TCCR1B = (1<<CS10);
    TIMSK |= (1<<OCIE1A);
    OCR1A += 10000;

    // INT0 init

    MCUCR |= (1<<ISC01) | (1<<ISC00);
    GIMSK |= (1<<INT0);
    EIFR   = (1<<INTF0);        // aktuellen  Interrupt lschen

    sei();                      // startet auch die Uhr

    // LCD init

    lcd_init();
    lcd_generatechar_P( 1, (PGM_P) user_char_0 );

    lcd_setcursor(0,0);
    lcd_string_P(PSTR(" Geigerz" "\xE1" "hler"));
    lcd_setcursor(0,1);
    lcd_string_P(PSTR(" Reloaded 2.0"));
    _delay_ms(5000);
    LCD_LIGHT_OFF

    // Konstantes Texte schreiben

    lcd_clear();
    lcd_string_P(PSTR("CPM     00:00:00"));
    lcd_setcursor(0,1);
    lcd_string_P(PSTR("  \x01     \xF6"));

    // Hauptschleife

    while(1) {
        set_sleep_mode(SLEEP_MODE_IDLE);
        sleep_mode();

        if (flag_100ms) {
            flag_100ms=0;

            // atomar globale Daten kopieren
            cli();
            m_time_ss = time_ss;
            m_time_mm = time_mm;
            m_time_hh = time_hh;
            m_cpm = cpm;
            m_ca = ca >> 2;
            m_ct = ct;                                   
            sei();

            // LCD neu schreiben, nur die Zahlen
            
            // CPM, Pulse pro Minute
            my_uitoa(m_cpm, tmp_string);
            lcd_setcursor(4,0);
            //lcd_string(tmp_string+7);     // fhrende Nullen
            my_print_number(tmp_string+7);  // ohne

            // CT, absolute Anzahl Pulse, Exponentialdarstellung mit berlauferkennung
            
            lcd_setcursor(10,1);
            
            if (m_ct == -1L) {               // berlauf
                lcd_string_P(PSTR("-OVL- "));
            } else if (m_ct > 99999999) {   // Einheit 1M
                my_uitoa(m_ct / 1000000, tmp_string);
                my_print_number(tmp_string+5);
                lcd_data('M');
            } else if (m_ct > 999999) {     // Einheit 1k
                my_uitoa(m_ct / 1000, tmp_string);
                my_print_number(tmp_string+5);
                lcd_data('k');
            } else {                        // Einheit 1
                my_uitoa(m_ct , tmp_string);
                my_print_number(tmp_string+4);
            }

            // CA, Mittelwert der letzten 4 Minuten
            my_uitoa(m_ca, tmp_string);
            lcd_setcursor(4,1);
            //lcd_string(tmp_string+7);     // fhrende Nullen
            my_print_number(tmp_string+7);  // ohne

            // Uhrzeit
            lcd_setcursor(8,0);

            print_time(m_time_hh);
            lcd_data(':');

            print_time(m_time_mm);
            lcd_data(':');

            print_time(m_time_ss);
            
        }
    }
}

// Timer 1 ISR, 10ms

ISR (TIMER1_COMPA_vect) {

    static uint8_t cnt_1s, cnt_100ms, cnt_light;
    static uint8_t avg_index;

    OCR1A += 10000;     // freilaufender Timer 1

    // Uhr

    cnt_100ms++;
    if(cnt_100ms == 10) {       // 100ms
        cnt_100ms=0;
        flag_100ms=1;
        
        if (cnt_light) {
            cnt_light--;
            if (cnt_light ==0) {
                LCD_LIGHT_OFF
            };
        }
    }

    cnt_1s++;
    if (cnt_1s == 100) {        // 1 Sekunde
        cnt_1s=0;

        time_ss++;
        if (time_ss == 60) {    // 1 Minute
            time_ss=0;
        
            // Mittelwert und CPM reset

            ca = ca - mittelwerte[avg_index] +cpm;
            mittelwerte[avg_index] = cpm;
            cpm=0;
            avg_index = (avg_index+1) & 0x3;

            time_mm++;
            if (time_mm == 60) {    // 1 Stunde
                time_mm=0;
                time_hh++;
                if (time_hh == 24) { // 1 Tag
                    time_hh=0;
                }
            }
        }
    }

    // Monoflop fr Beleuchtung

    if (KEY_PRESSED) {
        cnt_light=LIGHT_TIMEOUT;
        LCD_LIGHT_ON
    }
}

// steigende Flanke vom Puls detektieren

ISR (INT0_vect) {
    
    cpm++;
    if (ct != -1) ct++;
    
    BEEP_ON
    _delay_us(BEEP_TIME);
    BEEP_OFF
}

